class DataSDKError(Exception):
    """Base exception class."""


class DataFrameParsingError(DataSDKError):
    """Unable to parse DataFrame"""


class UnknownMediaTypeError(DataSDKError):
    """Unknown Media Type"""


class UnknownDatasetNameError(DataSDKError):
    """Unknown Dataset Name"""


class AuthenticationError(DataSDKError):
    """AuthenticationError"""


CREDENTIALS_NOT_WRITABLE_MSG = """\
Credentials directory not writable.
Either make $HOME/.config/unfolded writable or supply another credentials_dir.
"""

REFRESH_TOKEN_SAVED_MSG = """\
Your refresh token has been successfully saved to disk.
Future interactions with the Data SDK on this machine and user via Python or
via the CLI do not need to pass in authentication parameters manually.
"""

NON_ROTATING_TOKEN_WARNING = """\
Warning: a non-rotating refresh token will never expire. Take care to keep this token safe.
"""



import base64
import hashlib
import secrets
import socket
import webbrowser
from contextlib import closing
from http import HTTPStatus
from http.server import BaseHTTPRequestHandler, HTTPServer
from typing import Any, Optional, Sequence
from urllib.parse import parse_qs, urlencode, urlparse

import requests
from pydantic import parse_obj_as

from unfolded.data_sdk.environment import AUTH_CONFIGS, AuthEnvironment
from unfolded.data_sdk.errors import AuthenticationError
from unfolded.data_sdk.models import Credentials
from unfolded.data_sdk.utils import raise_for_status

# Whitelisted ports for a redirect_uri on localhost
REDIRECT_URI_PORT_RANGE = range(8050, 8054)


def authenticate(
    *,
    auth_env: AuthEnvironment = AuthEnvironment.PRODUCTION,
) -> Credentials:
    auth_config = AUTH_CONFIGS[auth_env]

    nonce = secrets.token_urlsafe(32)
    code_verifier = secrets.token_urlsafe(32)

    # The last = needs to be stripped off
    # https://www.oauth.com/oauth2-servers/pkce/authorization-request/
    code_challenge = base64.urlsafe_b64encode(
        hashlib.sha256(code_verifier.encode()).digest()
    ).rstrip(b"=")

    # Choose hostname and port
    hostname = "localhost"
    port = find_available_port(hostname, REDIRECT_URI_PORT_RANGE)
    redirect_uri = f"http://{hostname}:{port}"

    authorization_params = {
        "response_type": "code",
        "response_mode": "query",
        "audience": auth_config.audience,
        "state": nonce,
        "code_challenge": code_challenge,
        "code_challenge_method": "S256",
        "client_id": auth_config.client_id,
        "redirect_uri": redirect_uri,
        "scope": auth_config.scope,
    }

    authorization_url = f"{auth_config.auth_url}/authorize?" + urlencode(
        authorization_params
    )
    webbrowser.open(authorization_url)

    # Receive token from successful flow
    code = receive_token(hostname=hostname, port=port, nonce=nonce)

    refresh_token_params = {
        "grant_type": "authorization_code",
        "audience": auth_config.audience,
        "client_id": auth_config.client_id,
        "code_verifier": code_verifier,
        "code": code,
        "redirect_uri": redirect_uri,
    }

    refresh_token_url = f"{auth_config.auth_url}/oauth/token"
    r = requests.post(refresh_token_url, json=refresh_token_params)
    raise_for_status(r)
    return parse_obj_as(Credentials, r.json())


def receive_token(
    *, port: int, hostname: str = "localhost", global_timeout: int = 600, nonce: str
) -> str:
    """Set up HTTPServer on localhost to receive code and state query params from OAuth2

    Kwargs:
        hostname (str): hostname for server to accept "code". Defaults to "localhost".
        port (int): port for HTTPServer.
        global_timeout (int, optional): Timeout (in seconds) beyond which the Python server shuts down. Defaults to 600.

    Returns:
        Tuple[str, Optional[str]]: code and state
    """
    code: Optional[str] = None

    class Handler(BaseHTTPRequestHandler):
        """Custom request handler to intercept GET query parameters"""

        def do_GET(self) -> None:
            query = urlparse(self.path).query
            parsed = parse_qs(query)

            # The error is expected to be one of the enum values at the bottom of this url:
            # https://www.oauth.com/oauth2-servers/authorization/the-authorization-response/
            error = parsed.get("error", [""])[0]
            if error:
                self.send_response(HTTPStatus.FORBIDDEN)
                self.send_header("Content-type", "text/html")
                self.end_headers()

                message = f"Error message: {error}"
                self.wfile.write(message.encode())
                return

            nonlocal code
            code = parsed.get("code", [""])[0]
            state = parsed.get("state", [""])[0]

            if not state or state != nonce:
                self.send_response(HTTPStatus.BAD_REQUEST)
                self.send_header("Content-type", "text/html")
                self.end_headers()

                message = f"Internal error: state was missing or not equal to nonce ({nonce})."
                self.wfile.write(message.encode())
                return

            if not code:
                self.send_response(HTTPStatus.BAD_REQUEST)
                self.send_header("Content-type", "text/html")
                self.end_headers()

                message = "Internal error: code was not sent from authorize page."
                self.wfile.write(message.encode())
                return

            self.send_response(HTTPStatus.OK)
            self.send_header("Content-type", "text/html")
            self.end_headers()

            message = "Success! You may now close this page."
            self.wfile.write(message.encode())

        # Turn off default printing of HTTP requests
        def log_request(
            self, *args: Any, **kwargs: Any  # pylint: disable=unused-argument
        ) -> None:
            return

    with HTTPServer((hostname, port), Handler) as httpd:
        # NOTE(kyle): I _think_ we only need to call `handle_request` once, rather than in a loop,
        # and it will automatically time out after `timeout`
        # https://docs.python.org/3/library/socketserver.html#socketserver.BaseServer.timeout
        httpd.timeout = global_timeout
        httpd.handle_request()

    if not code:
        raise AuthenticationError(
            "Timeout while waiting for user to finish authentication."
        )

    return code


def find_available_port(host: str, port_range: Sequence[int]) -> int:
    """Find available/open port in range on specified host"""
    for port in port_range:
        if check_socket(host, port):
            return port

    raise OSError(f"No port in range {port_range} is available")


def check_socket(host: str, port: int) -> bool:
    """Check if port is available

    https://stackoverflow.com/a/35370008

    Returns True if port is available, otherwise False
    """
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        return sock.connect_ex((host, port)) != 0



"""Stub file to allow CLI use as an entry point

E.g.

```
python -m unfolded.data_sdk
```

instead of

```
uf-data-sdk
```

This can be useful to ensure that the CLI is using the same Python environment as the
active one. Otherwise in some circumstances, the PATH could be set up so that the
`uf-data-sdk` CLI and `python` are two different environments.
"""
from unfolded.data_sdk.cli import main

if __name__ == "__main__":
    main()



from __future__ import annotations

import gzip
import os
import shutil
import zlib
from contextlib import contextmanager
from typing import IO, Any, Callable, Generator, Iterable, Optional

import requests
from requests.exceptions import HTTPError
from tqdm import tqdm

from unfolded.data_sdk.errors import AuthenticationError, DataSDKError


@contextmanager
def keep_file_position(fileobj: IO[bytes]) -> Generator[None, None, None]:
    """Operate on a file but return to original position"""
    pos = fileobj.tell()
    try:
        yield
    finally:
        fileobj.seek(pos)


def get_fileobj_length(fileobj: IO) -> int:
    """Get length of file object"""
    with keep_file_position(fileobj):
        fileobj.seek(0, os.SEEK_END)
        length = fileobj.tell()
    return length


def read_fileobj_chunks(
    fileobj: IO[bytes], chunk_size: int, callback: Optional[Callable[[int], Any]] = None
) -> Iterable[bytes]:
    """Generator to read a file object by chunks"""
    while True:
        data = fileobj.read(chunk_size)
        if not data:
            break

        if callback:
            callback(len(data))

        yield data


def is_gzipped(fileobj: IO[bytes]) -> bool:
    """Check if file is gzipped by reading magic bytes"""
    with keep_file_position(fileobj):
        fileobj.seek(0)
        is_gzip = fileobj.read(2) == b"\x1f\x8b"
    return is_gzip


# from https://stackoverflow.com/a/24342024
def get_gzip_length(fileobj: IO[bytes]) -> int:
    """Estimate uncompressed size of gzipped file.
    Does not write to disk, instead serially decompresses
    the gzip into a fixed length in-memory buffer.
    """

    READ_CHUNK_SIZE = 1024 * 1024
    MAX_DECOMPRESS_SIZE = 4 * 1024 * 1024

    total_length = 0

    with keep_file_position(fileobj):
        buffer = fileobj.read(READ_CHUNK_SIZE)
        while buffer != b"":  # loop through concatenated gzip streams
            # 15: use a 2**15 size window, 16: expect gzip
            # see https://docs.python.org/3/library/zlib.html#zlib.decompress
            z = zlib.decompressobj(15 + 16)
            while not z.eof:  # loop through one gzip stream
                if buffer == b"":
                    raise ValueError("Incomplete gzip file")
                while buffer != b"":  # go through all output from one input buffer
                    total_length += len(z.decompress(buffer, MAX_DECOMPRESS_SIZE))
                    buffer = z.unconsumed_tail
                buffer = fileobj.read(READ_CHUNK_SIZE)
            buffer = z.unused_data  # get beginning of next gzip file
            if buffer == b"":
                buffer == fileobj.read(READ_CHUNK_SIZE)

        return total_length


# creates progress bar
def create_progress_bar(total_size: int, description: str, fout: IO) -> tqdm:
    return tqdm(
        total=total_size, desc=description, file=fout, unit="B", unit_scale=True
    )


# BinaryIO is supposed to be an alias for IO[bytes], but for some reason this
# fails with BinaryIO? Seems related to https://stackoverflow.com/q/62745734 but
# that's supposed to be fixed already.
def compress_fileobj(file_in: IO[bytes], file_out: IO[bytes], **kwargs: Any) -> None:
    """Apply gzip compression to file object

    Args:
        file_in: readable file object (in binary mode) with data to be compressed.
        file_out: writable file object (in binary mode) where compressed data should be written.
        **kwargs: keyword arguments to pass to gzip.open
    """
    with gzip.open(file_out, "wb", **kwargs) as gzipf:
        shutil.copyfileobj(file_in, gzipf)


def raise_for_status(r: requests.Response) -> None:
    """Check valid response, raising custom error for invalid authorization"""
    try:
        r.raise_for_status()
    except HTTPError as e:
        # Re-raise authentication error with better error message
        if r.status_code in (401, 403):
            raise AuthenticationError("Invalid Access Token") from e

        msg: str | None = None
        try:
            msg = r.json()["message"]
        except Exception:
            pass
        if msg is not None:
            raise DataSDKError(msg) from e

        raise e



import sys
from abc import ABC
from enum import Enum
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

from pydantic import BaseModel, Field

from unfolded.data_sdk.enums import NodeType
from unfolded.data_sdk.models import TimeInterval

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class QueryNode(BaseModel, ABC):
    """Abstract base class for query nodes"""

    type: NodeType
    input: List["QueryNode"] = Field(default_factory=list)

    class Config:
        """Allows using aliases to set attributes (which we don't set, but prevents against
        validation errors where pydantic complains about the alias not set when the attribute is
        set using the model name)
        https://pydantic-docs.helpmanual.io/usage/model_config/
        """

        allow_population_by_field_name = True


class DatasetNode(QueryNode):
    """A QueryNode to represent FROM <Dataset>"""

    type: NodeType = NodeType.DATASET
    input: List[QueryNode] = Field(default_factory=list, min_items=0, max_items=0)
    uuid: UUID


class InlineDataFormat(str, Enum):
    """Format for which the InlineData is sent"""

    CSV = "CSV"
    JSON_RECORDS = "JSON_OBJECT"


class InlineDataNode(QueryNode):
    """A QueryNode to represent FROM <Inline Data>"""

    type: NodeType = NodeType.INLINE_DATA
    input: List[QueryNode] = Field(default_factory=list, min_items=0, max_items=0)
    format: InlineDataFormat
    data: Union[List[Dict], str]


class SelectNode(QueryNode):
    """A QueryNode to represent SELECT <COLUMN>, <EXPRESSION>"""

    type: NodeType = NodeType.SELECT
    input: List[QueryNode] = Field(min_items=1, max_items=1)
    expressions: List[str] = Field(min_items=1)


class FilterNode(QueryNode):
    """A QueryNode to represent WHERE <CONDITION>"""

    type: NodeType = NodeType.FILTER
    input: List[QueryNode] = Field(min_items=1, max_items=1)
    expression: str


class GroupNode(QueryNode):
    """A QueryNode to represent GROUP BY <COLUMN>"""

    type: NodeType = NodeType.GROUP
    input: List[QueryNode] = Field(min_items=1, max_items=1)
    by: List[str] = Field(min_items=1)
    expressions: List[str] = Field(min_items=1)


class JoinType(str, Enum):
    LEFT = "LEFT"
    INNER = "INNER"
    FULL = "FULL"


class JoinExpression(BaseModel):
    columns: List[str]


class JoinNode(QueryNode):
    """A QueryNode to represent <Table1> LEFT JOIN <Table2> ON <CONDITION>"""

    type: NodeType = NodeType.JOIN
    input: List[QueryNode] = Field(min_items=2, max_items=2)
    on: List[JoinExpression] = Field(min_items=1)
    join: JoinType


class LimitNode(QueryNode):
    """A QueryNode to represent LIMIT <OFFSET>, <LENGTH>"""

    type: NodeType = NodeType.LIMIT
    input: List[QueryNode] = Field(min_items=1, max_items=1)
    length: int = Field(gt=0)
    offset: int = 0


class SortNode(QueryNode):
    """A QueryNode to represent ORDER BY <COLUMN> <ASC|DESC>"""

    type: NodeType = NodeType.SORT
    input: List[QueryNode] = Field(min_items=1, max_items=1)
    by: List[str] = Field(min_items=1)


class EnrichmentTargetType(str, Enum):
    """Enum for types of columns used for enrichment"""

    H3 = "H3"
    LATLNG = "LATLNG"


class EnrichNode(QueryNode):
    type: NodeType = NodeType.ENRICH
    input: List[QueryNode] = Field(min_items=1, max_items=1)
    source_id: UUID = Field(..., alias="sourceId")
    source_column: Union[str, List[str]] = Field(..., alias="sourceColumn")
    time_column: str = Field(None, alias="timeColumn")


class H3EnrichNode(EnrichNode):
    """Node for enriching using an H3 column"""

    target_type: Literal[EnrichmentTargetType.H3] = Field(
        EnrichmentTargetType.H3, alias="targetType"
    )
    column: str


class LatLngEnrichNode(EnrichNode):
    """Node for enriching using latitude and longitude columns"""

    target_type: Literal[EnrichmentTargetType.LATLNG] = Field(
        EnrichmentTargetType.LATLNG, alias="targetType"
    )
    lat_column: str = Field(..., alias="latColumn")
    lng_column: str = Field(..., alias="lngColumn")


class HexifyNode(QueryNode):
    """Node for hexifying a GeoJSON at a given resolution"""

    type: NodeType = NodeType.HEXIFY
    geojson: Dict
    res: int
    column_name: Optional[str] = Field(None, alias="columnName")


class TileExtractNode(QueryNode):
    """Node for extracting data from a HexTile dataset"""

    type: NodeType = NodeType.TILE_EXTRACT
    source_id: UUID = Field(..., alias="sourceId")
    source_column: Optional[Union[str, List[str]]] = Field(None, alias="sourceColumn")
    geojson: Dict
    res: Optional[int] = None
    h3_column_name: Optional[str] = Field(None, alias="h3ColumnName")
    time_column_name: Optional[str] = Field(None, alias="timeColumnName")
    time_interval: Optional[TimeInterval] = Field(None, alias="timeInterval")


class Query:
    """Query provides methods that can be chained to build an Unfolded query, and work with it
    seamlessly

    Example:
        q = Query() \
            .dataset(uuid='1da42194-a676-4e3c-99da-dca399905f11') \
            .filter(expression='metric > 100') \
            .group(by=["country"], expressions=['sum(metric1)', 'max(metric2)']) \
            .select(expressions=['country', 'metric1', 'metric1 * metric2']) \
            .sort(by=['desc(metric1)']) \
            .limit(length=10) \
            .json()

    """

    def __init__(self, node: Optional[QueryNode] = None):
        self.node = node

    def dataset(self, uuid: Union[UUID, str]) -> "Query":
        """Specifies the dataset source for which to apply the query

        Note:
            dataset() needs to be called as the first method in a query method chain, and cannot
            be applied on top of already chained query.

        Args:
            uuid: the dataset id (Unfolded dataset).

        Returns
            A Query object that can be chained with other query operators.
        """
        if self.node:
            raise RuntimeError(
                "dataset() cannot be chained to an existing query method"
            )
        node = DatasetNode(uuid=uuid)
        return Query(node)

    def inline_data(
        self,
        data: Union[List[Dict], str],
        *,
        # pylint:disable = redefined-builtin
        format: InlineDataFormat = InlineDataFormat.CSV,
    ) -> "Query":
        """Specifies the data to run the query with inline

        Args:
            data: the inline data, either as a string or a list of flat dictionaries (record array).

        Returns:
            A query object that can be chained with other query methods.
        """
        if self.node:
            raise RuntimeError("data() cannot be chained to an existing query method")

        node = InlineDataNode(format=format, data=data)
        return Query(node)

    def select(self, expressions: List[str]) -> "Query":
        """Selects columns or expression from a dataset

        Args:
            expressions: the list of expressions or columns to select from the dataset.

        Returns:
            A Query object that can be chained with other query methods.
        """
        node = SelectNode(expressions=expressions, input=[self.node])
        return Query(node)

    def filter(self, expression: str) -> "Query":
        """Applies a filter expression

        Args:
            expression: the expression to use to filter the dataset rows.

        Returns:
            A Query object that can be chained with other query methods.
        """
        node = FilterNode(expression=expression, input=[self.node])
        return Query(node)

    def group(self, by: List[str], expressions: List[str]) -> "Query":
        """Applies a group by aggregation

        Args:
            by: the columns to group by.
            expressions: the expressions to apply on the grouped values.

        Returns:
            A Query object that can be chained with other query methods.
        """
        node = GroupNode(by=by, expressions=expressions, input=[self.node])
        return Query(node)

    def sort(self, by: List[str]) -> "Query":
        """Sorts the dataset rows given a list of expressions

        Args:
            by: the list of sort expressions

        Returns:
            A Query object that can be chained with other query methods.
        """
        node = SortNode(by=by, input=[self.node])
        return Query(node)

    def limit(self, length: int, *, offset: int = 0) -> "Query":
        """Limits the number of rows returned by the query result

        Args:
            length: the number of rows to return.
            offset: the offset from which to limit the rows returned.

        Returns:
            A Query object that can be chained with other query methods.
        """
        node = LimitNode(length=length, offset=offset, input=[self.node])
        return Query(node)

    def enrich(
        self,
        source_id: Union[str, UUID],
        source_column: Union[str, List[str]],
        *,
        h3_column: Optional[str] = None,
        lat_column: Optional[str] = None,
        lng_column: Optional[str] = None,
        time_column: Optional[str] = None,
    ) -> "Query":
        """Enriches a dataset with a given enrichment source

        Enriches a dataset with a given enrichment source and column based on either a h3_column
        or a combination of lat_column and lng_column.

        Args:
            source_id: the uuid of the EnrichmentDataset to enrich with.
            source_column: the EnrichmentColumn to enrich with.
            h3_column: the dataset h3 column to use for enrichment (enrichment by h3).
            lat_column: the dataset latitude column to use for enrichment (enrichment by lat/lng).
            lng_column: the dataset longitude column to use for enrichment (enrichment by lat/lng).
            time_column: the dataset time column to use for temporal enrichment.

        Returns:
            A Query object that can be chained with other query methods.
        """

        node: Optional[EnrichNode] = None

        if h3_column:
            node = H3EnrichNode(
                source_id=source_id,
                source_column=source_column,
                column=h3_column,
                time_column=time_column,
                input=[self.node],
            )
        else:
            node = LatLngEnrichNode(
                source_id=source_id,
                source_column=source_column,
                lat_column=lat_column,
                lng_column=lng_column,
                time_column=time_column,
                input=[self.node],
            )
        return Query(node)

    def hexify(
        self, geojson: Union[Dict, str], res: int, column_name: Optional[str] = None
    ) -> "Query":
        """Polyfills a GeoJSON into H3 hexagons at a given resolution

        Args:
            geojson: the GeoJSON feature to polyfill.
            res: the H3 resolution used for hexification.
            column_name: the name of the column used in the resulting table.

        Returns:
            A Query object that can be chained with other query methods.
        """
        node = HexifyNode(geojson=geojson, res=res, column_name=column_name)
        return Query(node)

    def tile_extract(
        self,
        source_id: Union[str, UUID],
        geojson: Dict,
        *,
        source_column: Optional[Union[str, List[str]]] = None,
        res: Optional[int] = None,
        h3_column: Optional[str] = None,
        time_column: Optional[str] = None,
        time_interval: Optional[Union[Dict, TimeInterval]] = None,
    ) -> "Query":
        """

        Returns:
            A Query object that can be chained with other query methods.
        """
        node = TileExtractNode(
            source_id=source_id,
            source_column=source_column,
            geojson=geojson,
            res=res,
            h3_column_name=h3_column,
            time_column_name=time_column,
            time_interval=time_interval,
        )
        return Query(node)

    def dict(self, **kwargs: Any) -> Dict:
        """Returns a python dictionary representation of the query

        Returns:
            A python dictionary representing the query tree
        """
        if not self.node:
            raise RuntimeError("Cannot convert an empty query to a dictionary")

        return self.node.dict(**kwargs)

    def json(self, **kwargs: Any) -> str:
        """Returns a json representation of the query

        Returns:
            A json string representing the query tree
        """
        if not self.node:
            raise RuntimeError("Cannot convert an empty query to json")

        kwargs["exclude_none"] = True
        kwargs["by_alias"] = True

        return self.node.json(**kwargs)



from enum import Enum


class MediaType(str, Enum):
    CSV = "text/csv"
    GEOJSON = "application/geo+json"
    JSON = "application/json"
    FGB = "application/x.flatgeobuf"
    PMTILES = "application/vnd.pmtiles"


class QueryOutputType(str, Enum):
    CSV = "csv"
    PARQUET = "parquet"


class DatasetType(str, Enum):
    MANAGED = "managed"
    EXTERNALLY_HOSTED = "externally-hosted"
    VECTOR_TILE = "vector-tile"
    RASTER_TILE = "raster-tile"
    HEX_TILE = "hex-tile"
    SQL = "sql"
    WMS = "wms"
    TILE3D = "tile3d"


class DataConnectorType(str, Enum):
    SNOWFLAKE = "snowflake"
    BIG_QUERY = "big-query"
    ATHENA = "athena"
    POSTGRES = "postgres"
    PRESTO = "presto"
    REDSHIFT = "redshift"
    PLANET = "planet"
    S3 = "s3"
    GOOGLE_CLOUD_STORAGE = "google-cloud-storage"
    DATABRICKS = "databricks"


class HexTileFieldType(str, Enum):
    """Possible field types for hextile datasets"""

    STRING = "string"
    INTEGER = "integer"
    REAL = "real"
    BOOLEAN = "boolean"
    DATE = "date"
    TIMESTAMP = "timestamp"
    GEOJSON = "geojson"
    POINT = "point"


class AggregationMethod(str, Enum):
    """Available aggregation types
    Values should be strings or functions that can be passed to pd.DataFrame.agg
    """

    SUM = "sum"
    COUNT = "count"
    MIN = "min"
    MAX = "max"
    MEAN = "mean"
    MEDIAN = "median"
    MODE = "mode"


class Dtype(str, Enum):
    """Available data types"""

    INT64 = "int64"
    INT32 = "int32"
    INT16 = "int16"
    INT8 = "int8"
    UINT64 = "uint64"
    UINT32 = "uint32"
    UINT16 = "uint16"
    UINT8 = "uint8"
    BOOL = "bool"
    FLOAT64 = "float64"
    FLOAT32 = "float32"
    FLOAT16 = "float16"


class TimeInterval(str, Enum):
    """Time granularities for hextile job input."""

    YEAR = "YEAR"
    MONTH = "MONTH"
    DAY = "DAY"
    HOUR = "HOUR"
    MINUTE = "MINUTE"
    SECOND = "SECOND"


class TileMode(str, Enum):
    """Available tile modes"""

    DENSE = "dense"
    SPARSE = "sparse"
    AUTO = "auto"


class JobSize(str, Enum):
    """Available sizes for hex tile jobs"""

    S = "S"
    M = "M"
    L = "L"
    XL = "XL"


class NodeType(str, Enum):
    """Enum for the supported QueryNode types"""

    DATASET = "dataset"
    INLINE_DATA = "inline_data"
    SELECT = "select"
    GROUP = "group"
    FILTER = "filter"
    SORT = "sort"
    LIMIT = "limit"
    SAMPLE = "sample"
    JOIN = "join"
    ENRICH = "enrich"
    HEXIFY = "hexify"
    TILE_EXTRACT = "tileExtract"


class ResourceType(str, Enum):
    """Resource types"""

    MAP = "map"
    DATASET = "dataset"
    # we coerce this type into an actual `data-connection` value
    DATA_CONNECTOR = "data-connector"


class PermissionType(str, Enum):
    """Permission types"""

    VIEWER = "viewer"
    EDITOR = "editor"



from __future__ import annotations

import json
import sys
from functools import partial
from getpass import getpass
from pathlib import Path
from typing import IO, Any, List

import click
from pydantic.json import pydantic_encoder

from unfolded.data_sdk.auth import authenticate as _authenticate
from unfolded.data_sdk.data_sdk import DataSDK, RefreshToken
from unfolded.data_sdk.enums import ResourceType
from unfolded.data_sdk.environment import AuthEnvironment
from unfolded.data_sdk.errors import NON_ROTATING_TOKEN_WARNING, REFRESH_TOKEN_SAVED_MSG
from unfolded.data_sdk.models import CategorizedPermissions, MediaType


class PathType(click.Path):
    """A Click path argument that returns a pathlib Path, not a string"""

    def convert(self, value: Any, param: Any, ctx: Any) -> Any:
        return Path(super().convert(value, param, ctx))


@click.group()
def main() -> None:
    pass


@click.command()
@click.option(
    "--credentials-dir",
    type=PathType(dir_okay=True, file_okay=False, writable=True),
    help="A directory on disk to use for storing credentials. Only used if --store-credentials is passed.",
    default=Path("~/.config/unfolded/").expanduser(),
    show_default=True,
)
@click.option(
    "--non-rotating",
    is_flag=True,
    help="If passed, generate a non-rotating refresh token.",
    default=False,
    show_default=True,
)
@click.option(
    "--print/--no-print",
    "_print",
    default=False,
    show_default=True,
    help="If passed, print refresh token information to stdout.",
)
@click.option(
    "--store-credentials/--no-store-credentials",
    default=True,
    show_default=True,
    help="If passed, store generated credentials on disk to the location of --credentials-dir.",
)
def authenticate(
    credentials_dir: Path, non_rotating: bool, _print: bool, store_credentials: bool
) -> None:
    """Authenticate with the Unfolded backend using OAuth2"""
    auth_env = (
        AuthEnvironment.PRODUCTION_NO_ROTATE
        if non_rotating
        else AuthEnvironment.PRODUCTION
    )

    if non_rotating:
        sys.stderr.write("\n\x1b[1;33m" + NON_ROTATING_TOKEN_WARNING + "\x1b[0m\n")

    credentials = _authenticate(auth_env=auth_env)

    if store_credentials:
        data_sdk = DataSDK(
            access_token=credentials.access_token,
            credentials_dir=credentials_dir,
            store_credentials=store_credentials,
        )
        data_sdk.refresh_token = credentials.refresh_token
        sys.stderr.write(REFRESH_TOKEN_SAVED_MSG)

    if _print:
        click.echo(json.dumps(credentials, indent=4, default=pydantic_encoder))


@click.command()
@click.option(
    "--organization/--no-organization",
    default=False,
    help="If True, list datasets for organization of authenticated user. For non-enterprise users, passing this flag will cause the request to fail with a 403 error",
    show_default=True,
)
def list_datasets(organization: bool) -> None:
    """List datasets for authenticated user"""
    data_sdk = DataSDK()
    output_data = [
        dataset.dict() for dataset in data_sdk.list_datasets(organization=organization)
    ]
    click.echo(json.dumps(output_data, indent=4, default=pydantic_encoder))


@click.command()
@click.option("--dataset-id", type=str, required=True, help="Dataset id.")
def get_dataset(dataset_id: str) -> None:
    """Get dataset given its id"""
    data_sdk = DataSDK()
    dataset = data_sdk.get_dataset_by_id(dataset=dataset_id)
    click.echo(dataset.json(indent=4))


@click.command()
@click.option("--dataset-id", type=str, required=True, help="Dataset id.")
@click.option(
    "-o",
    "--output-file",
    type=PathType(file_okay=True, writable=True),
    required=True,
    help="Output file for dataset.",
)
@click.option(
    "--progress/--no-progress",
    default=True,
    help="Whether to show progress bar.",
    show_default=True,
)
def download_dataset(dataset_id: str, output_file: Path, progress: bool) -> None:
    """Download data for existing dataset to disk"""
    data_sdk = DataSDK()
    data_sdk.download_dataset(
        dataset=dataset_id, output_file=output_file, progress=progress
    )


@click.command()
@click.option(
    "-n", "--name", type=str, required=False, default=None, help="Dataset name."
)
@click.option(
    "--media-type",
    type=click.Choice([c.value for c in MediaType], case_sensitive=False),
    required=False,
    default=None,
    help="Dataset media type.",
)
@click.option(
    "--dataset-id",
    type=str,
    required=False,
    default=None,
    help="Dataset id. If provided, will update the existing dataset.",
)
@click.option(
    "--desc",
    type=str,
    required=False,
    default=None,
    show_default=True,
    help="Dataset description.",
)
@click.option(
    "--progress/--no-progress",
    default=True,
    help="Whether to show progress bar.",
    show_default=True,
)
@click.argument("file", type=PathType(readable=True, file_okay=True))
def upload_file(
    file: Path, name: str, media_type: str, dataset_id: str, desc: str, progress: bool
) -> None:
    """Upload new dataset to Unfolded Studio"""
    data_sdk = DataSDK()
    new_dataset = data_sdk.upload_file(
        file=file,
        name=name,
        media_type=media_type,
        description=desc,
        dataset=dataset_id,
        progress=progress,
    )
    click.echo(new_dataset.json())


@click.command()
@click.option("--dataset-id", type=str, required=True, help="The dataset id")
@click.option("--name", type=str, required=False, help="The new dataset name")
@click.option(
    "--description", type=str, required=False, help="The new dataset description"
)
@click.option(
    "--file",
    type=PathType(readable=True, file_okay=True),
    required=False,
    help="File to update dataset with",
)
@click.option(
    "--media-type",
    type=click.Choice([c.value for c in MediaType], case_sensitive=False),
    required=False,
    default=None,
    help="Dataset media type.",
)
@click.option(
    "--progress/--no-progress",
    default=True,
    help="Whether to show progress bar.",
    show_default=True,
)
def update_dataset(
    dataset_id: str,
    name: str,
    description: str,
    file: Path,
    media_type: str,
    progress: bool,
) -> None:
    """Update data for existing Unfolded Studio dataset"""
    data_sdk = DataSDK()
    updated_dataset = data_sdk.update_dataset(
        dataset_id=dataset_id,
        name=name,
        description=description,
        file=file,
        media_type=media_type,
        progress=progress,
    )
    click.echo(updated_dataset.json())


def abort_if_false(ctx: Any, param: Any, value: Any) -> None:
    # pylint: disable=unused-argument
    if not value:
        ctx.abort()


@click.command()
@click.option("--dataset-id", type=str, required=True, help="Dataset id.")
@click.option(
    "--force",
    is_flag=True,
    callback=abort_if_false,
    expose_value=False,
    help="Delete dataset without prompting.",
    prompt="Are you sure you want to delete the dataset?",
)
def delete_dataset(dataset_id: str) -> None:
    """Delete dataset from Unfolded Studio

    Warning: This operation cannot be undone. If you delete a dataset currently
    used in one or more maps, the dataset will be removed from those maps,
    possibly causing them to render incorrectly.
    """
    data_sdk = DataSDK()
    data_sdk.delete_dataset(dataset=dataset_id)
    click.echo("Dataset deleted.", file=sys.stderr)


@click.command()
@click.argument("map_id", type=str)
@click.argument("dataset-to-replace-id", type=str)
@click.argument("dataset-to-use-id", type=str)
@click.option(
    "--force",
    is_flag=True,
    default=False,
    help="Force replace dataset (even if not compatible)",
)
@click.option(
    "--strict",
    is_flag=True,
    default=True,
    help="Use strict type checking for datasets",
)
def replace_dataset(
    map_id: str,
    dataset_to_replace_id: str,
    dataset_to_use_id: str,
    force: bool = False,
    strict: bool = False,
) -> None:
    """Replace one dataset associated with a map with another"""
    data_sdk = DataSDK()
    map_object = data_sdk.replace_dataset(
        map_id,
        dataset_to_replace=dataset_to_replace_id,
        dataset_to_use=dataset_to_use_id,
        force=force,
        strict=strict,
    )
    click.echo(map_object.json())


@click.command()
@click.option(
    "--organization/--no-organization",
    default=False,
    help="If True, list map records for organization of authenticated user. For non-enterprise users, passing this flag will cause the request to fail with a 403 error",
    show_default=True,
)
def list_maps(organization: bool) -> None:
    """List map records for authenticated user"""
    data_sdk = DataSDK()
    output_data = [
        map.dict(exclude_none=True)
        for map in data_sdk.list_maps(organization=organization)
    ]
    click.echo(json.dumps(output_data, indent=4, default=pydantic_encoder))


@click.command()
@click.argument("map_id", type=str)
def get_map(map_id: str) -> None:
    """Get an Unfolded map record by its id, including associated datasets and map state"""
    data_sdk = DataSDK()
    map_object = data_sdk.get_map_by_id(map_id)
    click.echo(map_object.json())


@click.command()
@click.option("--name", type=str, required=True, help="The map name")
@click.option("--description", type=str, required=False, help="The map description")
@click.option(
    "--map-state",
    type=click.File(),
    required=False,
    help="The map state as a json file",
)
@click.option(
    "--dataset-ids",
    type=str,
    required=False,
    help="Comma separated list of dataset ids to add to the map",
)
def create_map(name: str, description: str, map_state: IO, dataset_ids: str) -> None:
    """Create an Unfolded Studio map"""
    data_sdk = DataSDK()
    datasets = (
        [dataset_id.strip() for dataset_id in dataset_ids.split(",")]
        if dataset_ids
        else None
    )

    map_object = data_sdk.create_map(
        name=name,
        description=description,
        map_state=json.load(map_state) if map_state else None,
        datasets=datasets,
    )

    click.echo(map_object.json())


@click.command()
@click.argument("map_id", type=str)
@click.option(
    "--copy-datasets/--no-copy-datasets",
    required=True,
    help="If True, will copy the map and copy its underlying datasets.",
)
@click.option(
    "--name",
    type=str,
    required=False,
    help="The name for the new copied map. Default: 'Copy of {source_map_name}'",
)
def copy_map(map_id: str, copy_datasets: bool, name: str) -> None:
    """Copy an Unfolded map"""
    data_sdk = DataSDK()
    map_object = data_sdk.copy_map(map_id, copy_datasets=copy_datasets, name=name)
    click.echo(map_object.json())


@click.command()
@click.option("--map-id", type=str, required=True, help="The map id")
@click.option("--name", type=str, required=False, help="The map name")
@click.option("--description", type=str, required=False, help="The map description")
@click.option(
    "--map-state",
    type=click.File(),
    required=False,
    help="The map state as a json file",
)
@click.option(
    "--dataset-ids",
    type=str,
    required=False,
    help="Comma separated list of dataset ids to add to the map",
)
def update_map(
    map_id: str, name: str, description: str, map_state: IO, dataset_ids: str
) -> None:
    """Update an Unfolded Studio map"""
    data_sdk = DataSDK()
    datasets = (
        [dataset_id.strip() for dataset_id in dataset_ids.split(",")]
        if dataset_ids
        else None
    )
    map_object = data_sdk.update_map(
        map_id=map_id,
        name=name,
        description=description,
        map_state=json.load(map_state) if map_state else None,
        datasets=datasets,
    )
    click.echo(map_object.json())


@click.command()
@click.option(
    "--force",
    is_flag=True,
    callback=abort_if_false,
    expose_value=False,
    help="Delete map without prompting.",
    prompt="Are you sure you want to delete this map?",
)
@click.argument("map_id", type=str)
def delete_map(map_id: str) -> None:
    """Delete map from Unfolded Studio

    Warning: This operation cannot be undone.
    """
    data_sdk = DataSDK()
    data_sdk.delete_map(map_id)
    click.echo("Map deleted.", file=sys.stderr)


@click.command()
@click.option(
    "--organization/--no-organization",
    default=False,
    help="If True, list data connectors for organization of authenticated user. For non-enterprise users, passing this flag will cause the request to fail with a 403 error",
    show_default=True,
)
def list_data_connectors(organization: bool) -> None:
    """List data connectors for authenticated user"""
    data_sdk = DataSDK()
    output_data = data_sdk.list_data_connectors(organization=organization)
    click.echo(json.dumps(output_data, indent=4, default=pydantic_encoder))


@click.command()
@click.option(
    "--connector-id", type=str, required=True, help="Id of Data Connector to use"
)
@click.option(
    "--query", type=str, required=True, help="SQL query to run against the connector"
)
@click.option("--output-file", type=str, help="The path to write the query output to")
@click.option(
    "--output-format", type=str, help="The format in which to write the output"
)
def execute_query(
    connector_id: str, query: str, output_file: str | None, output_format: str | None
) -> None:
    """Execute a query against a data connector (enterprise only)"""
    data_sdk = DataSDK()
    output_df = data_sdk.execute_query(
        connector=connector_id,
        query=query,
        output_file=output_file,
        output_format=output_format,
    )
    if output_file:
        click.echo(f"Query results written to: {output_file}")
    else:
        click.echo(repr(output_df))


@click.command()
@click.option(
    "--connector-id", type=str, required=True, help="Id of Data Connector to use"
)
@click.option("--query", type=str, required=True, help="SQL query to use")
@click.option(
    "--name", type=str, default=None, required=True, help="Name of the new SQL dataset"
)
@click.option(
    "--description",
    type=str,
    default=None,
    required=False,
    help="Description of the new SQL dataset",
)
def create_query_dataset(
    connector_id: str, query: str, name: str, description: str | None
) -> None:
    """Create a dataset from a query against a data connector (enterprise only)"""
    data_sdk = DataSDK()
    dataset = data_sdk.create_query_dataset(
        connector=connector_id, query=query, name=name, description=description
    )
    click.echo(dataset.json())


@click.command()
@click.option("--source", type=str, required=True, help="Source URL of the dataset")
@click.option(
    "--name", type=str, default=None, required=True, help="Name of the new dataset"
)
@click.option(
    "--description",
    type=str,
    default=None,
    required=False,
    help="Description of the new dataset",
)
@click.option(
    "--connector-id",
    type=str,
    required=False,
    help="Id of optional Data Connector to use",
)
def create_external_dataset(
    source: str, name: str, description: str | None, connector_id: str | None
) -> None:
    """Create a dataset from an external file"""
    data_sdk = DataSDK()
    dataset = data_sdk.create_external_dataset(
        connector=connector_id, source=source, name=name, description=description
    )
    click.echo(dataset.json())


@click.command()
@click.option("--source", type=str, required=True, help="Source dataset id.")
@click.option("--target", type=str, required=False, help="Target dataset id.")
@click.option(
    "--source-hex-column", type=str, required=False, help="Source hex column."
)
@click.option(
    "--source-lat-column", type=str, required=False, help="Source lat column."
)
@click.option(
    "--source-lng-column", type=str, required=False, help="Source lng column."
)
@click.option(
    "--source-time-column", type=str, required=False, help="Source time column."
)
@click.option(
    "--time-interval",
    type=str,
    multiple=True,
    required=False,
    help='Time granularities for hextile job input. Options are "YEAR", "MONTH", "DAY", "HOUR", "MINUTE", "SECOND". Supports multiple inputs',
)
@click.option(
    "--finest-resolution", type=int, required=False, help="Finest resolution."
)
@click.option(
    "--output-column",
    type=str,
    multiple=True,
    required=False,
    help="Output column config (JSON). Supports multiple inputs",
)
def generate_hextile(
    source: str,
    target: str | None,
    source_hex_column: str | None,
    source_lat_column: str | None,
    source_lng_column: str | None,
    source_time_column: str | None,
    time_interval: List[str],
    finest_resolution: int | None,
    output_column: List[str],
) -> None:
    """Generate a hextile dataset from a source dataset"""
    data_sdk = DataSDK()
    dataset = data_sdk.generate_hextile(
        source=source,
        target=target,
        source_hex_column=source_hex_column,
        source_lat_column=source_lat_column,
        source_lng_column=source_lng_column,
        source_time_column=source_time_column,
        time_intervals=list(time_interval) if time_interval else None,
        finest_resolution=finest_resolution,
        output_columns=list(map(json.loads, output_column)),
    )
    click.echo(dataset.json(indent=4))


@click.command()
@click.option("--source", type=str, required=True, help="Source dataset id.")
@click.option("--target", type=str, required=False, help="Target dataset id.")
@click.option(
    "--source-lat-column", type=str, required=False, help="Source lat column."
)
@click.option(
    "--source-lng-column", type=str, required=False, help="Source lng column."
)
@click.option(
    "-y",
    "--attributes",
    type=str,
    required=False,
    multiple=True,
    help="Attributes to keep.",
)
@click.option(
    "-X",
    "--exclude-all-attributes",
    type=str,
    required=False,
    is_flag=True,
    help="Whether to exclude all attributes.",
)
@click.option(
    "--tile-size-kb", type=int, required=False, help="Maximum tile size (in kilobytes)."
)
def generate_vectortile(
    source: str,
    target: str | None,
    source_lat_column: str | None,
    source_lng_column: str | None,
    attributes: List[str],
    exclude_all_attributes: bool,
    tile_size_kb: int | None,
) -> None:
    """Generate a vector tileset from a source dataset"""
    data_sdk = DataSDK()
    dataset = data_sdk.generate_vectortile(
        source=source,
        target=target,
        source_lat_column=source_lat_column,
        source_lng_column=source_lng_column,
        attributes=attributes,
        exclude_all_attributes=exclude_all_attributes,
        tile_size_kb=tile_size_kb,
    )
    click.echo(dataset.json(indent=4))


@click.command()
@click.option(
    "--refresh-token",
    type=str,
    help="Refresh Token. Retrieve from https://studio.foursquare.com/tokens.html",
    # Use getpass for password input
    # Ref https://github.com/pallets/click/issues/300#issuecomment-606105993
    default=partial(getpass, "Refresh Token: "),
)
@click.option(
    "--credentials-dir",
    type=PathType(dir_okay=True, file_okay=False, writable=True),
    help="A directory on disk to use for storing credentials.",
    default=Path("~/.config/unfolded/").expanduser(),
    show_default=True,
)
def store_refresh_token(refresh_token: RefreshToken, credentials_dir: Path) -> None:
    """Store refresh token to enable seamless future authentication

    Retrieve token from https://studio.foursquare.com/tokens.html
    """
    DataSDK(refresh_token=refresh_token, credentials_dir=credentials_dir)


@click.command()
@click.option(
    "--resource-type",
    type=click.Choice([c.value for c in ResourceType], case_sensitive=False),
    required=True,
    help="Resource type",
)
@click.option("--resource-id", type=str, required=True, help="Resource UUID")
def get_permissions(resource_type: ResourceType, resource_id: str) -> None:
    """Get permissions for a specific resource"""

    data_sdk = DataSDK()
    permissions = data_sdk.get_permissions(
        resource_type=resource_type, resource_id=resource_id
    )
    click.echo(permissions.json())


@click.command()
@click.option(
    "--resource-type",
    type=click.Choice([c.value for c in ResourceType], case_sensitive=False),
    required=True,
    help="Resource type",
)
@click.option("--resource-id", type=str, required=True, help="Resource UUID")
@click.option(
    "--organization",
    type=click.Choice(["viewer", "editor"]),
    required=False,
    help="Permissions for an organization",
)
@click.option(
    "viewers",
    "--viewer",
    type=str,
    required=False,
    multiple=True,
    help="Email address of user(s) who should have 'viewer' permissions",
)
@click.option(
    "editors",
    "--editor",
    type=str,
    required=False,
    multiple=True,
    help="Email address of user(s) who should have 'editor' permissions",
)
def set_permissions(
    resource_type: ResourceType,
    resource_id: str,
    organization: str | None,
    viewers: tuple[str],
    editors: tuple[str],
) -> None:
    """Set permissions for a specific resource"""

    data_sdk = DataSDK()

    user_permissions: list[dict] = []
    for viewer_email in viewers:
        user_permissions.append({"email": viewer_email, "permission": "viewer"})
    for editor_email in editors:
        user_permissions.append({"email": editor_email, "permission": "editor"})

    permissions = {"organization": organization, "users": user_permissions}

    data_sdk.set_permissions(
        resource_type=resource_type,
        resource_id=resource_id,
        permissions=CategorizedPermissions(**permissions),
    )


main.add_command(authenticate)
main.add_command(list_datasets)
main.add_command(get_dataset)
main.add_command(download_dataset)
main.add_command(upload_file)
main.add_command(update_dataset)
main.add_command(delete_dataset)
main.add_command(list_maps)
main.add_command(get_map)
main.add_command(create_map)
main.add_command(copy_map)
main.add_command(update_map)
main.add_command(delete_map)
main.add_command(list_data_connectors)
main.add_command(execute_query)
main.add_command(create_query_dataset)
main.add_command(create_external_dataset)
main.add_command(generate_hextile)
main.add_command(generate_vectortile)
main.add_command(store_refresh_token)
main.add_command(get_permissions)
main.add_command(set_permissions)

if __name__ == "__main__":
    main()



# Pydantic validators are class methods, so require the first param to be `cls`, but
# pylint can't recognize this: https://github.com/samuelcolvin/pydantic/issues/568
# pylint: disable=no-self-argument
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple, Union
from uuid import UUID

from pydantic import (
    BaseModel,
    Field,
    StrictBool,
    StrictInt,
    StrictStr,
    root_validator,
    validator,
)

from unfolded.data_sdk.enums import (
    AggregationMethod,
    DataConnectorType,
    DatasetType,
    Dtype,
    HexTileFieldType,
    JobSize,
    MediaType,
    PermissionType,
    ResourceType,
    TileMode,
    TimeInterval,
)
from unfolded.data_sdk.types import AccessToken, RefreshToken


class Credentials(BaseModel):
    access_token: AccessToken
    refresh_token: RefreshToken
    scope: str
    expires_in: int
    token_type: str


class DatasetMetadata(BaseModel):
    """A model representing metadata for an Unfolded Studio Dataset"""

    media_type: Optional[Union[MediaType, StrictStr]] = Field(alias="contentType")
    size: Optional[StrictInt]
    source: Optional[StrictStr]
    tileset_data_url: Optional[StrictStr] = Field(alias="tilesetDataUrl")
    tileset_metadata_url: Optional[StrictStr] = Field(alias="tilesetMetadataUrl")
    image_url: Optional[StrictStr] = Field(alias="imageUrl")
    metadata_url: Optional[StrictStr] = Field(alias="metadataUrl")
    data_status: Optional[StrictStr] = Field(alias="dataStatus")

    class Config:
        allow_population_by_field_name = True


class DatasetUpdateParams(BaseModel):
    """A model representing creation and update parameters for Datasets"""

    name: Optional[StrictStr]
    description: Optional[StrictStr]

    class Config:
        allow_population_by_field_name = True


class DataConnector(BaseModel):
    """A model representing a data connector"""

    id: UUID
    name: StrictStr
    description: Optional[StrictStr]
    created_at: datetime = Field(..., alias="createdAt")
    updated_at: datetime = Field(..., alias="updatedAt")
    type: DataConnectorType

    class Config:
        allow_population_by_field_name = True


class Dataset(BaseModel):
    """A model representing an Unfolded Studio Dataset"""

    id: UUID
    name: StrictStr
    type: DatasetType
    created_at: datetime = Field(..., alias="createdAt")
    updated_at: datetime = Field(..., alias="updatedAt")
    description: Optional[StrictStr]
    is_valid: StrictBool = Field(..., alias="isValid")
    data_connector: Optional[DataConnector] = Field(alias="dataConnection")
    metadata: DatasetMetadata

    class Config:
        allow_population_by_field_name = True


class MapState(BaseModel):
    """A model representing an Unfolded Studio Map Starte"""

    id: UUID
    # data contains the actual map configuration, and should be modeled more concretely than a
    # generic Dictionary.
    # Todo (wesam@unfolded.ai): revisit this once we have a style building strategy
    data: Dict

    class Config:
        allow_population_by_field_name = True


class MapUpdateParams(BaseModel):
    """A model respresenting creation and update parameters for Unfolded Maps"""

    name: Optional[StrictStr]
    description: Optional[StrictStr]
    latest_state: Optional[MapState] = Field(None, alias="latestState")
    datasets: Optional[List[UUID]]

    class Config:
        allow_population_by_field_name = True


class Map(BaseModel):
    """A model representing an Unfolded Studio Map"""

    id: UUID
    name: StrictStr
    description: Optional[StrictStr]
    created_at: datetime = Field(..., alias="createdAt")
    updated_at: datetime = Field(..., alias="updatedAt")
    latest_state: Optional[MapState] = Field(None, alias="latestState")
    datasets: Optional[List[Dataset]]

    class Config:
        allow_population_by_field_name = True


class ConnectorQuery(BaseModel):
    """A model representing a SQL query to execute or create a dataset from"""

    connector_id: UUID = Field(..., alias="connectionId")
    query: StrictStr

    class Config:
        allow_population_by_field_name = True


class HexTileMetadataField(BaseModel):
    """A model representing a field (column) within hextile dataset metadata"""

    name: StrictStr
    type: HexTileFieldType
    domain: Optional[Tuple[float, float]]

    class Config:
        allow_population_by_field_name = True
        # Ignore extra properties
        extra = "ignore"


class HexTileMetadata(BaseModel):
    """A model representing an metadata for a hextile dataset"""

    fields: List[HexTileMetadataField]
    name: Optional[StrictStr] = None
    min_zoom: StrictInt = Field(..., alias="minZoom")
    max_zoom: StrictInt = Field(..., alias="maxZoom")
    resolution_offset: StrictInt = Field(..., alias="resolutionOffset")
    version: StrictStr

    class Config:
        allow_population_by_field_name = True
        # Ignore extra properties
        extra = "ignore"


class HexTileOutputColumnConfig(BaseModel):
    source_column: StrictStr = Field(..., alias="sourceColumn")
    """Column name in source dataset"""

    target_column: StrictStr = Field(..., alias="targetColumn")
    """Column name in target dataset"""

    agg_method: Optional[AggregationMethod] = Field(..., alias="aggMethod")
    """Aggregation method to use for column"""

    dtype: Optional[Dtype] = None
    """Dtype to use for column"""

    class Config:
        allow_population_by_field_name = True


class H3Resolution(StrictInt):
    ge = 0
    le = 15


class TilingBaseModel(BaseModel):
    """Config for tiling process"""

    source: UUID
    """Source Dataset"""

    @validator("source", pre=True, always=True)
    def _coerce_source_to_uuid(cls, v: Any) -> Any:
        if isinstance(v, Dataset):
            return v.id

        return v

    target: Optional[UUID] = None
    """Target Dataset"""

    @validator("target", pre=True, always=True)
    def _coerce_target_to_uuid(cls, v: Any) -> Any:
        if isinstance(v, Dataset):
            return v.id

        return v


class HexTileConfig(TilingBaseModel):
    """Config for hextiling process"""

    @root_validator
    def _validate_source_columns(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        source_hex_column = values.get("source_hex_column")
        source_lat_column = values.get("source_lat_column")
        source_lng_column = values.get("source_lng_column")
        source_time_column = values.get("source_time_column")
        time_intervals = values.get("time_intervals")

        if source_hex_column:
            if source_lat_column or source_lng_column:
                raise ValueError(
                    "Only provide source_hex_column or source_lat_column and source_lng_column, but not all three"
                )
        else:
            if not source_lat_column or not source_lng_column:
                raise ValueError(
                    "Must provide either source_hex_column or source_lat_column and source_lng_column"
                )

        if time_intervals and not source_time_column:
            raise ValueError(
                "Must provide source_time_column when passing in time_interval"
            )

        return values

    source_hex_column: Optional[StrictStr] = Field(None, alias="sourceHexColumn")
    """Name of the hex column in the source dataset"""

    source_lat_column: Optional[StrictStr] = Field(None, alias="sourceLatColumn")
    """Name of the lat column in the source dataset"""

    source_lng_column: Optional[StrictStr] = Field(None, alias="sourceLngColumn")
    """Name of the lng column in the source dataset"""

    source_time_column: Optional[StrictStr] = Field(None, alias="sourceTimeColumn")
    """Name of the time column, in the source dataset"""

    time_intervals: Optional[List[TimeInterval]] = Field(
        None, alias="timeIntervals", min_items=1
    )
    """List of time intervals to use for temporal datasets"""

    target_res_offset: Optional[H3Resolution] = Field(None, alias="targetResOffset")
    """
    Offset between the resolution of the tile to the resolution of the data within it
    """

    finest_resolution: Optional[H3Resolution] = Field(None, alias="finestResolution")
    """
    Finest resolution for the data hexes within a tile (when creating a tileset from lat/lng columns)
    """

    experimental_tile_mode: Optional[TileMode] = Field(None, alias="tileMode")

    output_columns: Optional[List[HexTileOutputColumnConfig]] = Field(
        None, alias="outputColumns", min_items=1
    )

    experimental_positional_indexes: Optional[StrictBool] = Field(
        None, alias="positionalIndexes"
    )

    job_size: Optional[JobSize] = Field(None, alias="jobSize")

    class Config:
        allow_population_by_field_name = True


class VectorTileConfig(TilingBaseModel):

    source_lat_column: Optional[StrictStr] = Field(None, alias="sourceLatColumn")
    """Name of the lat column in the source dataset"""

    source_lng_column: Optional[StrictStr] = Field(None, alias="sourceLngColumn")
    """Name of the lng column in the source dataset"""

    attributes: Optional[List[StrictStr]]
    """List of attributes to keep in vector tiling. Leave blank to keep all."""

    exclude_all_attributes: Optional[bool] = Field(None, alias="excludeAllAttributes")
    """Whether to exclude all attributes in vector tiling."""

    tile_size_kb: Optional[int] = Field(None, alias="tileSizeKb")
    """Maximum tile size (in kilobytes) for each generated tile."""


class UserPermission(BaseModel):
    email: StrictStr
    permission: PermissionType


class CategorizedPermissions(BaseModel):
    organization: Optional[PermissionType]
    users: Optional[List[UserPermission]]


class PermissionsConfig(BaseModel):
    resource_type: Union[ResourceType, str] = Field(..., alias="resourceType")
    resource_id: Union[str, UUID] = Field(..., alias="resourceId")
    permissions: Union[CategorizedPermissions, Dict]

    class Config:
        use_enum_values = True
        allow_population_by_field_name = True



from enum import Enum

from pydantic import BaseModel


class EnvironmentConfig(BaseModel):
    """EnvironmentConfig specifies request and authentication endpoints for a given environment"""

    base_url: str = "https://data-api.foursquare.com"
    studio_url: str = "https://studio.foursqure.com"


class AuthEnvironment(str, Enum):
    """Enum for the different Unfolded environments"""

    PRODUCTION = "production"
    PRODUCTION_NO_ROTATE = "production-no-rotate"


class AuthConfig(BaseModel):
    client_id: str
    auth_url: str = "https://auth.studio.foursquare.com"
    audience: str = "https://foursquare.com/api/"
    scope: str = "offline_access"


AUTH_CONFIGS = {
    AuthEnvironment.PRODUCTION: AuthConfig(
        client_id="v970dpbcqmRtr3y9XwlAB3dycpsvNRZF",
    ),
    AuthEnvironment.PRODUCTION_NO_ROTATE: AuthConfig(
        client_id="75nXztDByx9A3525kDyTAqDeLSMWEwyj",
    ),
}

PRODUCTION_ENV = EnvironmentConfig()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['data_sdk']

package_data = \
{'': ['*']}

install_requires = \
['PyJWT>=2.0.0,<3.0.0',
 'click>=7.0,<9.0',
 'pydantic>=1.7.0,<2.0.0',
 'requests>=2.31.0,<3.0.0',
 'tqdm>=4.64.0,<5.0.0']

extras_require = \
{':python_version < "3.8"': ['typing-extensions>=3.7.4,<4.0.0'],
 'dataframe': ['pandas>=1.0,<2.0'],
 'dataframe:python_version >= "3.8"': ['geopandas>=0.13.2'],
 'query': ['pandas>=1.0,<2.0', 'pyarrow>=6.0.0,<7.0.0'],
 'test': ['pandas>=1.0,<2.0', 'pyarrow>=6.0.0,<7.0.0'],
 'test:python_version >= "3.8"': ['geopandas>=0.13.2']}

entry_points = \
{'console_scripts': ['uf-data-sdk = unfolded.data_sdk.cli:main']}

setup_kwargs = {
    'name': 'unfolded.data-sdk',
    'version': '0.17.1',
    'description': "Module for working with Foursquare Studio's Data SDK",
    'long_description': "# `unfolded.data-sdk`\n\nPython package for interfacing with [Unfolded](https://unfolded.ai)'s Data SDK.\n\nFor more documentation, refer to the [documentation website](https://location.foursquare.com/developer/docs/studio-data-sdk-overview).\n",
    'author': 'Kyle Barron',
    'author_email': 'kbarron@foursquare.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)
